/*
 * Anonymous artifact for CHES (major revision), submission deadline July 15, 2022 AoE
 */

#include <math.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <immintrin.h>
#include <time.h>
#include "int32_sort.h"

// defines for trace control
#define TRACELEN 1//10784
#define EVAL_START 0//10536
#define EVAL_END (EVAL_START+1)
#define EVALLEN (EVAL_END - EVAL_START)
#define TRACES_PER_FILE 1000
#define TRACES_NUM 100000000LU
#define EVAL_NUM    1500000LU

#define MAX_ITERATIONS 66
#define EVOLUTION_RESOLUTION 100 // number of traces that are processed before a new ranking is stored

// control coordinates within the state which are attacked
#define MINX 0 // 0-4
#define MAXX 5 // 1-5
#define MINZ 0 // 0-63
#define MAXZ 64 // 1-64
#define NUMX (MAXX-MINX)
#define NUMZ (MAXZ-MINZ)

#define ATTACK_BIT 5 // number of bits in the key
#define ATTACK_SPACE (1<<ATTACK_BIT) // number of potential solutions
#define M_BIT 5 // number of bits in M
#define M_SPACE (1<<M_BIT) // number of potential values for M

// helper macros for indexing
#define I(k, kappa, m, mu) (((m)<<(M_BIT+2*ATTACK_BIT))|((kappa)<<(2*ATTACK_BIT))|((k)<<(ATTACK_BIT))|(mu))
#define GETBIT(x, pos) (((x)>>((pos)%ATTACK_BIT))&1)
#define GETCOLUMN(arr, z, x) (((((arr)[x*5 + 4]>>z)&1) << 0) |((((arr)[x*5 + 3]>>z)&1) << 1) |((((arr)[x*5 + 2]>>z)&1) << 2) |((((arr)[x*5 + 1]>>z)&1) << 3) |((((arr)[x*5 + 0]>>z)&1) << 4))
#define BIT(x, pos) ((((x)[pos/8])>>((pos)%8))&1)

// helper macros for allocating memory
#define CALLOC8(x, len) do {x = calloc(len, sizeof(uint8_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC16(x, len) do {x = calloc(len, sizeof(int16_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC32(x, len) do {x = calloc(len, sizeof(int32_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC64(x, len) do {x = calloc(len, sizeof(int64_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)

// uninitialized reference vector
int16_t ref[ATTACK_SPACE * ATTACK_SPACE * ATTACK_SPACE * M_SPACE];

// precomputation for reference vector
void precompute(void)
{
  uint32_t k,m,mu,kappa;
  // iterate over all k, kappa, m, mu
  for (kappa = 0; kappa < ATTACK_SPACE; kappa++)
  {
    for (m = 0; m < M_SPACE; m++)
    {
      for (mu = 0; mu < ATTACK_SPACE; mu++)
      {
        for (k = 0; k < ATTACK_SPACE; k++)
        {
          int16_t sum = 0;
          // iterate over all bits to be attacked
          for (int i = 0; i < ATTACK_BIT; i++)
          {
            // activation function
            uint8_t tmp = GETBIT(k, i+2) ^ GETBIT(kappa, i+2) ^ GETBIT(m, i+2) ^ GETBIT(mu, i+2) 
                        ^ ((GETBIT(kappa, i+1) ^ GETBIT(mu, i+1) ^ 1) & (GETBIT(kappa, i) ^ GETBIT(mu, i)));
            if (tmp == 0)
            {
              sum += 1;
            } else {
              sum -= 1;
            }
          }
          ref[I(k, kappa, m, mu)] = sum;
        }
      } 
    }
  }
}

// correlation coefficient helper function
double corrcoef(int64_t Sxy, int64_t Sx, int64_t Sx2, int64_t Sy, int64_t Sy2, uint64_t n)
{
  int64_t numerator = n*Sxy - (Sx * Sy);
  int64_t X = n*Sx2 - Sx*Sx;
  int64_t Y = n*Sy2 - Sy*Sy;
  double denominator = sqrt(X) * sqrt(Y);
  return (double)numerator / denominator;
}

int main(void)
{
  FILE *f;
  char fname[128];
  
  // known k and kappa for the measurement
  uint64_t correct_k[] = {
    0x44d08a5cdc6a367, 0x2843458764905a3, 0xd2d9950cf2bb4056, 0xb4f1c565ed386d1e, 0x8e66d1ddb9499c22, 
    0xaa456744c9238c0c, 0xea8e320172dd8a31, 0xa66549b247ae0812, 0x384853cebcd4dd5d, 0x745eebee50395cd2, 
    0x7d2a93cd5a111883, 0x350c1eb3dda02e7e, 0xe0e7d63179b26e92, 0x76e7b7c5e9b2fea0, 0x1200963557d93978, 
    0xcf9e3ca6494f3002, 0xf933cfb932c27ca3, 0xc4d99957a6f69d9c, 0xe6676652dc4a6aea, 0x9939e773e2469f60, 
    0x82a3e4a8124632f4, 0x97d1438f1ced027d, 0xfafcdd4932217f5, 0x2fe97976c1919188, 0x6caa4779957593a9
  };
  uint64_t correct_kappa[] = {
    0x7e2f2d10ea6be843, 0x7e651cacbaa3a85c, 0x16d511a519f3c62d, 0x8f6c1859af04384, 0xc5dcec2930558d1, 
    0x877e2b23d08a837, 0xde0fe46836e7bb09, 0x3a45b3c3ede29d38, 0x2187b2ecaf058bb, 0xbb7d64b3a5e65478, 
    0x458358c702a6d386, 0x88687efeefaf089c, 0xa107e9713a2539e3, 0xd7917d4c3940dc96, 0x474d7ade743e307, 
    0x98b9b21cf40c453e, 0xe8ec728d02742f1e, 0x261a20aae9087855, 0xb88f0d7314d6084e, 0xb54b923f63abd7d7, 
    0x81a75f5153aafd19, 0xb02e198aedeb494d, 0xd373620ef7701dfc, 0xc4f64d6bf83226dd, 0xde536ed3afddb876
  };
  uint8_t ck[64][5] = {0};
  uint8_t ckappa[64][5] = {0};
  for (size_t z = MINZ; z < MAXZ; z++)
  {
    for (size_t x = MINX; x < MAXX; x++)
    {
      ck[z][x] = GETCOLUMN(correct_k, z, x);
      ckappa[z][x] = GETCOLUMN(correct_kappa, z, x);
    } 
  }
  // 0,  1,  3,  7, 27, 63
  ck[ 0][0] ^= 0x10; // correct for iota since we attack the second round
  ck[ 1][0] ^= 0x10;
  ck[ 3][0] ^= 0x10;
  ck[ 7][0] ^= 0x10;
  ck[27][0] ^= 0x10;
  ck[63][0] ^= 0x10;
  
  // precompute reference vectors
  precompute();
  
  size_t file_offset = 0;
  /////////////////////////////////////////////////////////////////////////////////////// loop over the independent iterations
  for (size_t iterations = 0; iterations < MAX_ITERATIONS; iterations++)
  {
    int8_t *traces;
    uint64_t *mu,*pt;
    int32_t *Sx, *Sy;
    int32_t *Sxy;
    int64_t *Sx2;
    int32_t *Sy2;
    uint16_t *rank;
    // allocate memory
    CALLOC16(rank, EVAL_NUM/EVOLUTION_RESOLUTION*NUMX*NUMZ);
    CALLOC32(Sx, EVALLEN);
    CALLOC32(Sy, ATTACK_SPACE*ATTACK_SPACE*NUMX*NUMZ);
    CALLOC32(Sxy, ATTACK_SPACE*ATTACK_SPACE*EVALLEN*NUMX*NUMZ);
    CALLOC64(Sx2, EVALLEN);
    CALLOC32(Sy2, ATTACK_SPACE*ATTACK_SPACE*NUMX*NUMZ);
    CALLOC8(traces, TRACELEN*TRACES_PER_FILE);
    CALLOC64(mu, 25*TRACES_PER_FILE);
    CALLOC64(pt, 25*TRACES_PER_FILE);
    
    
    ///////////////////////////////////////////////////////////////////////////////////// iteration over trace files
    size_t count_eval = 0;
    for (size_t i = file_offset; i < TRACES_NUM/TRACES_PER_FILE; i++)
    {
      srand(time(NULL));
      
      printf("%4lu / %4d iterations, %10lu / %10lu trace files\r", iterations, MAX_ITERATIONS, i, TRACES_NUM/TRACES_PER_FILE);
      fflush(stdout);
      
      // read traces
      snprintf(fname, 128, "traces/poi%d/poi0_%d.dat", (int)i/10000, (int)i);
      f = fopen(fname, "r");
      if (f == NULL || fread(traces, sizeof(int8_t), TRACELEN*TRACES_PER_FILE, f) != TRACELEN*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      // read mu
      snprintf(fname, 128, "plaintexts/plain%d/mu_%d.dat", (int)i/10000, (int)i);
      f = fopen(fname, "r");
      if (f == NULL || fread(mu, sizeof(uint64_t), 25*TRACES_PER_FILE, f) != 25*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      // read plaintexts
      snprintf(fname, 128, "plaintexts/plain%d/m_%d.dat", (int)i/10000, (int)i);
      f = fopen(fname, "r");
      if (f == NULL || fread(pt, sizeof(uint64_t), 25*TRACES_PER_FILE, f) != 25*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      /////////////////////////////////////////////////////////////////////////////////// iteration over traces in one file
      for (size_t j = 0; j < TRACES_PER_FILE; j++)
      {
        for (size_t sample = EVAL_START; sample < EVAL_END; sample++) // iteration over the included data points
        {
          // store sum of points and sum of squared points
          Sx[sample-EVAL_START] += traces[TRACELEN * j + sample];
          Sx2[sample-EVAL_START] += traces[TRACELEN * j + sample] * traces[TRACELEN * j + sample];
        }
  #pragma omp parallel for collapse (2)
        for (int x = MINX; x < MAXX; x++) // iteration over the attack rows
        {
          for (int z = MINZ; z < MAXZ; z++)
          {
            uint64_t thismu = GETCOLUMN(&mu[25*j], z, x);
            uint64_t thispt = GETCOLUMN(&pt[25*j], z, x);
  #pragma omp parallel for collapse (2)
            for (size_t kappa = 0; kappa < ATTACK_SPACE; kappa++) // each potential kappa
            {
              for (size_t k = 0; k < ATTACK_SPACE; k++) // each potential k
              {
                int oracle = ref[I(k, kappa, thispt, thismu)];
                // sum of oracle outputs and sum of their squares
                Sy[(((x - MINX)*NUMZ + z - MINZ)*ATTACK_SPACE + k) * ATTACK_SPACE + kappa] += oracle;
                Sy2[(((x - MINX)*NUMZ + z - MINZ)*ATTACK_SPACE + k) * ATTACK_SPACE + kappa] += oracle*oracle;
                for (size_t sample = EVAL_START; sample < EVAL_END; sample++) // each data point in the current trace
                {
                  // sum of oracle times data point
                  Sxy[((((x - MINX)*NUMZ + z - MINZ)*ATTACK_SPACE + k) * ATTACK_SPACE + kappa) * EVALLEN + sample - EVAL_START] += traces[TRACELEN * j + sample] * oracle;
                }
              } // k
            } // kappa 
          } // z
        } // x
        count_eval += 1;
        
        // if evolution resolution is reached, do a ranking snap shot
        // for each attacked row and each sample
        if ((count_eval%EVOLUTION_RESOLUTION) == 0)
        {
  #pragma omp parallel for collapse (2)
          for (int x = MINX; x < MAXX; x++)
          {
            for (int z = MINZ; z < MAXZ; z++)
            {
              for (size_t sample = 0; sample < EVALLEN; sample++)
              {
                double tvals[ATTACK_SPACE*ATTACK_SPACE] = {0}, maxabst = 0;
                int32_t sorthelper[ATTACK_SPACE*ATTACK_SPACE] = {0};
                for (size_t kappa = 0; kappa < ATTACK_SPACE; kappa++) // each potential kappa
                {
                  for (size_t k = 0; k < ATTACK_SPACE; k++) // each potential k
                  {
                    // compute correlation coefficient from the sums
                    tvals[k * ATTACK_SPACE + kappa] = corrcoef(Sxy[((((x - MINX)*NUMZ + z - MINZ)*ATTACK_SPACE + k) * ATTACK_SPACE + kappa) * EVALLEN + sample], 
                                    Sx[sample],
                                    Sx2[sample],
                                    Sy[(((x - MINX)*NUMZ + z - MINZ)*ATTACK_SPACE + k) * ATTACK_SPACE + kappa],
                                    Sy2[(((x - MINX)*NUMZ + z - MINZ)*ATTACK_SPACE + k) * ATTACK_SPACE + kappa],
				    count_eval);
                    // find maximum absolute correlation value
                    if (tvals[k * ATTACK_SPACE + kappa] > 0 && tvals[k * ATTACK_SPACE + kappa] > maxabst)
                    {
                      maxabst = tvals[k * ATTACK_SPACE + kappa];
                    }
                    if (tvals[k * ATTACK_SPACE + kappa] < 0 && -tvals[k * ATTACK_SPACE + kappa] > maxabst)
                    {
                      maxabst = -tvals[k * ATTACK_SPACE + kappa];
                    }
                  }
                }
                // pack the data into the sort helper (k, kappa and on top the scaled corrcoef such that sorting is for the corrcoef)
                for (size_t kappa = 0; kappa < ATTACK_SPACE; kappa++) // each potential kappa
                {
                  for (size_t k = 0; k < ATTACK_SPACE; k++) // each potential k
                  {
                    sorthelper[k * ATTACK_SPACE + kappa] = ((int32_t)(tvals[k * ATTACK_SPACE + kappa]*(((double)(1<<14))/maxabst))) << 16;
                    sorthelper[k * ATTACK_SPACE + kappa] &= 0xffff0000;
                    sorthelper[k * ATTACK_SPACE + kappa] |= k << 8;
                    sorthelper[k * ATTACK_SPACE + kappa] |= kappa;
                  }
                }
                int32_sort(sorthelper, ATTACK_SPACE*ATTACK_SPACE); // sort
                for (size_t h = 0; h < ATTACK_SPACE*ATTACK_SPACE; h++) // find the (known) correct k, kappa tuple and store its position (=rank)
                {
                  if ((sorthelper[h]&0xffff) == ((((int32_t)ck[z][x])<<8) | (int32_t)(ckappa[z][x])))
                  {
                    if (h > rank[((x - MINX)*NUMZ + z - MINZ) * (EVAL_NUM / EVOLUTION_RESOLUTION) + count_eval / EVOLUTION_RESOLUTION - 1])
                    {
                      rank[((x - MINX)*NUMZ + z - MINZ) * (EVAL_NUM / EVOLUTION_RESOLUTION) + count_eval / EVOLUTION_RESOLUTION - 1] = h;
                    }
                  }
                } // h
              } // sample
            } // z
          } // x
        } // fi
        
        if (count_eval >= EVAL_NUM)
        {
          break;
        }
      } // end iteration over traces
      if (count_eval >= EVAL_NUM)
      {
        file_offset = i+1;
        break;
      }
    } // end iteration over trace files
        
    // write the ranks
    snprintf(fname, 128, "keccak_cparanks_%03lu", iterations);
    f = fopen(fname, "w");
    fwrite(rank, sizeof(uint16_t), EVAL_NUM/EVOLUTION_RESOLUTION*NUMX*NUMZ, f);
    fclose(f);
  
        
    free(traces);
    free(mu);
    free(pt);
    free(Sx);
    free(Sy);
    free(Sx2);
    free(Sy2);
    free(Sxy);
    free(rank);
  } // iterations
}
