/*
 * CHES Artifact for the paper "A Tale of Snakes and Horses: Amplifying Correlation Power Analysis on Quadratic Maps" 
 * (TCHES 2024/1: https://tches.iacr.org/index.php/TCHES/article/view/11239).
 *
 * This program performs the snake attack on Keccak.
 */

#include <math.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <immintrin.h>
#include <time.h>
#include "int32_sort.h"

// defines for trace control
#define TRACELEN 1
#define EVAL_START 0
#define EVAL_END (EVAL_START+1)
#define EVALLEN (EVAL_END - EVAL_START)
#define TRACES_PER_FILE 1000
#define TRACES_NUM 100000000LU
#define EVAL_NUM    1500000LU

#define MAX_ITERATIONS 66
#define EVOLUTION_RESOLUTION 100 // number of traces that are processed before a new ranking is stored

// control coordinates within the state which are attacked
#define MINX 0 // 0-4
#define MAXX 5 // 1-5
#define MINZ 0 // 0-63
#define MAXZ 64 // 1-64
#define NUMX (MAXX-MINX)
#define NUMZ (MAXZ-MINZ)

#define ATTACK_BIT 5 // number of bits in the key
#define ATTACK_SPACE (1<<ATTACK_BIT) // number of potential solutions
#define M_BIT 5 // number of bits in M
#define M_SPACE (1<<M_BIT) // number of potential values for M
#define MU_BIT 3 // same for mu
#define MU_SPACE (1<<MU_BIT)
#define KAPPA_BIT 2 // same for kappa
#define KAPPA_SPACE (1<<KAPPA_BIT)

// helper macros for indexing
#define I(k, kappa, m, mu) (((m)<<(M_BIT+2*ATTACK_BIT))|((kappa)<<(2*ATTACK_BIT))|((k)<<(ATTACK_BIT))|(mu))
#define GETBIT(x, pos, mod) (((x)>>((pos)%(mod)))&1)
#define GETCOLUMN(arr, z, x) (((((arr)[x*5 + 4]>>z)&1) << 0) |((((arr)[x*5 + 3]>>z)&1) << 1) |((((arr)[x*5 + 2]>>z)&1) << 2) |((((arr)[x*5 + 1]>>z)&1) << 3) |((((arr)[x*5 + 0]>>z)&1) << 4))
#define BIT(x, pos) ((((x)[pos/8])>>((pos)%8))&1)

// helper macros for allocating memory
#define CALLOC8(x, len) do {x = calloc(len, sizeof(uint8_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC16(x, len) do {x = calloc(len, sizeof(int16_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC32(x, len) do {x = calloc(len, sizeof(int32_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC64(x, len) do {x = calloc(len, sizeof(int64_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)

// indices for the snake walk
size_t sumi[32][5] = {{0, 0, 0, 0, 0},{0, 0, 1, 1, 0},{0, 1, 2, 0, 0},{0, 1, 3, 1, 0},{1, 2, 0, 0, 0},{1, 2, 1, 1, 0},{1, 3, 2, 0, 0},{1, 3, 3, 1, 0},{2, 0, 0, 0, 1},{2, 0, 1, 1, 1},{2, 1, 2, 0, 1},{2, 1, 3, 1, 1},{3, 2, 0, 0, 1},{3, 2, 1, 1, 1},{3, 3, 2, 0, 1},{3, 3, 3, 1, 1},{0, 0, 0, 2, 2},{0, 0, 1, 3, 2},{0, 1, 2, 2, 2},{0, 1, 3, 3, 2},{1, 2, 0, 2, 2},{1, 2, 1, 3, 2},{1, 3, 2, 2, 2},{1, 3, 3, 3, 2},{2, 0, 0, 2, 3},{2, 0, 1, 3, 3},{2, 1, 2, 2, 3},{2, 1, 3, 3, 3},{3, 2, 0, 2, 3},{3, 2, 1, 3, 3},{3, 3, 2, 2, 3},{3, 3, 3, 3, 3}};

// signal reference vector
int8_t sref[5][KAPPA_SPACE][MU_SPACE] = {{{1, -1, 1, 1, -1, 1, -1, -1},
{-1, 1, 1, 1, 1, -1, -1, -1}, 
{1, 1, 1, -1, -1, -1, -1, 1}, 
{1, 1, -1, 1, -1, -1, 1, -1}}, 

{{1, -1, 1, 1, -1, 1, -1, -1,},
{-1, 1, 1, 1, 1, -1, -1, -1}, 
{1, 1, 1, -1, -1, -1, -1, 1}, 
{1, 1, -1, 1, -1, -1, 1, -1,}}, 

{{1, -1, 1, 1, -1, 1, -1, -1},
{-1, 1, 1, 1, 1, -1, -1, -1}, 
{1, 1, 1, -1, -1, -1, -1, 1}, 
{1, 1, -1, 1, -1, -1, 1, -1,}}, 

{{1, 1, -1, -1, -1, 1, 1, -1},
{1, 1, -1, -1, 1, -1, -1, 1}, 
{-1, 1, 1, -1,  1, 1, -1, -1}, 
{1, -1, -1, 1, 1, 1, -1, -1}}, 

{{1, -1, -1, 1,  1, -1, 1, -1},
{-1, 1, 1, -1, 1, -1, 1, -1}, 
{1, -1, 1, -1, 1, -1, -1, 1}, 
{1, -1, 1, -1, -1, 1, 1, -1}}};

// correllation coefficient helper function
double corrcoef(int64_t Sxy, int64_t Sx, int64_t Sx2, int64_t Sy, int64_t Sy2, uint32_t num)
{
  int64_t numerator = num*Sxy - (Sx * Sy);
  int64_t X = num*Sx2 - Sx*Sx;
  int64_t Y = num*Sy2 - Sy*Sy;
  double denominator = sqrt(X) * sqrt(Y);
  double result = (double)numerator / denominator;
  if (result < -1 || result > 1)
  {
    #pragma omp critical 
    printf("\nOver- or underflow detected.\n");
  }
  return result;
}

int main(void)
{
  FILE *f;
  char fname[128];
  
  // known kappa for the measurement
  uint64_t correct_kappa[] = {
    0x7e2f2d10ea6be843, 0x7e651cacbaa3a85c, 0x16d511a519f3c62d, 0x8f6c1859af04384, 0xc5dcec2930558d1, 
    0x877e2b23d08a837, 0xde0fe46836e7bb09, 0x3a45b3c3ede29d38, 0x2187b2ecaf058bb, 0xbb7d64b3a5e65478, 
    0x458358c702a6d386, 0x88687efeefaf089c, 0xa107e9713a2539e3, 0xd7917d4c3940dc96, 0x474d7ade743e307, 
    0x98b9b21cf40c453e, 0xe8ec728d02742f1e, 0x261a20aae9087855, 0xb88f0d7314d6084e, 0xb54b923f63abd7d7, 
    0x81a75f5153aafd19, 0xb02e198aedeb494d, 0xd373620ef7701dfc, 0xc4f64d6bf83226dd, 0xde536ed3afddb876
  };
  uint8_t ckappa[64][5] = {0};
  for (size_t z = MINZ; z < MAXZ; z++)
  {
    for (size_t x = MINX; x < MAXX; x++)
    {
      ckappa[z][x] = GETCOLUMN(correct_kappa, z, x);
    } 
  }
  

  size_t file_offset = 0;
  /////////////////////////////////////////////////////////////////////////////////////// loop over the independent iterations
  for (size_t iterations = 0; iterations < MAX_ITERATIONS; iterations++)
  {
    int8_t *traces;
    uint64_t *mu,*pt;
    int32_t *Sx, *Sy;
    int32_t *Sxy;
    int64_t *Sx2;
    int32_t *Sy2;
    uint16_t *rank;
    CALLOC16(rank, EVAL_NUM/EVOLUTION_RESOLUTION*NUMX*NUMZ);
    CALLOC32(Sx, EVALLEN);
    CALLOC32(Sy, 5*KAPPA_SPACE*NUMX*NUMZ);
    CALLOC32(Sxy, 5*KAPPA_SPACE*EVALLEN*NUMX*NUMZ);
    CALLOC64(Sx2, EVALLEN);
    CALLOC32(Sy2, 5*KAPPA_SPACE*NUMX*NUMZ);
    CALLOC8(traces, TRACELEN*TRACES_PER_FILE);
    CALLOC64(mu, 25*TRACES_PER_FILE);
    CALLOC64(pt, 25*TRACES_PER_FILE);
    
    
    
    ///////////////////////////////////////////////////////////////////////////////////// iteration over trace files
    size_t count_eval = 0;
    for (size_t it = file_offset; it < TRACES_NUM/TRACES_PER_FILE; it++)
    {
      srand(time(NULL));
      
      printf("%4lu / %4d iterations, %10lu / %10lu trace files\r", iterations, MAX_ITERATIONS, it, TRACES_NUM/TRACES_PER_FILE);
      fflush(stdout);
      
      // read traces
      snprintf(fname, 128, "traces/poi%d/poi0_%d.dat", (int)it/10000, (int)it);
      f = fopen(fname, "r");
      if (f == NULL || fread(traces, sizeof(int8_t), TRACELEN*TRACES_PER_FILE, f) != TRACELEN*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      // read mu
      snprintf(fname, 128, "plaintexts/plain%d/mu_%d.dat", (int)it/10000, (int)it);
      f = fopen(fname, "r");
      if (f == NULL || fread(mu, sizeof(uint64_t), 25*TRACES_PER_FILE, f) != 25*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      // read plaintexts
      snprintf(fname, 128, "plaintexts/plain%d/m_%d.dat", (int)it/10000, (int)it);
      f = fopen(fname, "r");
      if (f == NULL || fread(pt, sizeof(uint64_t), 25*TRACES_PER_FILE, f) != 25*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      /////////////////////////////////////////////////////////////////////////////////// iteration over traces in one file
      for (size_t j = 0; j < TRACES_PER_FILE; j++)
      {
        for (size_t sample = EVAL_START; sample < EVAL_END; sample++) // iteration over the included data points
        {
          // store sum of points and sum of squared points
          Sx[sample-EVAL_START] += traces[TRACELEN * j + sample];
          Sx2[sample-EVAL_START] += traces[TRACELEN * j + sample] * traces[TRACELEN * j + sample];
        }
  #pragma omp parallel for collapse (2)
        for (int x = MINX; x < MAXX; x++) // iteration over the attack rows
        {
          for (int z = MINZ; z < MAXZ; z++)
          {
            uint64_t thismu = GETCOLUMN(&mu[25*j], z, x);
            uint64_t thispt = GETCOLUMN(&pt[25*j], z, x);
  #pragma omp parallel for collapse (2)
            for (size_t kappa = 0; kappa < KAPPA_SPACE; kappa++) // each potential kappa
            {
              for (size_t i = 0; i < 5; i++) // iterate over the bits
              {
                uint64_t mutmp = 0;
                switch (i) { // only select the relevant bits from mu (reduced sref)
                  case 0: mutmp = thismu & 7;
                  case 1: mutmp = (thismu >> 1) & 7;
                  case 2: mutmp = thismu >> 2;
                  case 3: mutmp = (thismu >> 3) | ((thismu&1) << 2);
                  case 4: mutmp = (thismu >> 4) | ((thismu&3) << 1);
                  default:
                }
                int oracle = sref[i][kappa][mutmp];
                if (((thispt >> (4-i))&1))
                {
                  oracle *= -1;
                }
                // sum of oracle outputs and of their squares
                Sy[(((x - MINX)*NUMZ + z - MINZ)*5 + i) * KAPPA_SPACE + kappa] += oracle;
                Sy2[(((x - MINX)*NUMZ + z - MINZ)*5 + i) * KAPPA_SPACE + kappa] += oracle*oracle;
                for (size_t sample = EVAL_START; sample < EVAL_END; sample++) // each data point in the current trace
                {
                  // sum of oracle times data point
                  Sxy[((((x - MINX)*NUMZ + z - MINZ)*5 + i) * KAPPA_SPACE + kappa) * EVALLEN + sample - EVAL_START] += traces[TRACELEN * j + sample] * oracle;
                }
              } // i
            } // kappa
          } // z
        } // x
        count_eval += 1;
        
        // if evolution resolution is reached, do a ranking snap shot
        // for each attacked row and each sample
        if ((count_eval%EVOLUTION_RESOLUTION) == 0)
        {
  #pragma omp parallel for collapse (3)
          for (int x = MINX; x < MAXX; x++)
          {
            for (int z = MINZ; z < MAXZ; z++)
            {
              for (size_t sample = 0; sample < EVALLEN; sample++)
              {
                double tvals[5*KAPPA_SPACE] = {0}, sumtvals[KAPPA_SPACE*8], maxabst = -8*KAPPA_SPACE;
                int32_t sorthelper[KAPPA_SPACE*8] = {0};
                for (size_t kappa = 0; kappa < KAPPA_SPACE; kappa++) // each potential kappa
                {
                  for (size_t i = 0; i < 5; i++) 
                  {
                    // compute correlation coefficient from the sums
                    tvals[i * KAPPA_SPACE + kappa] = corrcoef(Sxy[((((x - MINX)*NUMZ + z - MINZ)*5 + i) * KAPPA_SPACE + kappa) * EVALLEN + sample], 
                                    Sx[sample],
                                    Sx2[sample],
                                    Sy[(((x - MINX)*NUMZ + z - MINZ)*5 + i) * KAPPA_SPACE + kappa],
                                    Sy2[(((x - MINX)*NUMZ + z - MINZ)*5 + i) * KAPPA_SPACE + kappa],
				    count_eval);
                  }
                }
                // sum the squared correlation coefficients according to the snake walk
                for (size_t kappa = 0; kappa < KAPPA_SPACE*8; kappa++)
                {
                  sumtvals[kappa] = 0;
                  for (size_t i = 0; i < 5; i++)
                  {
                    sumtvals[kappa] += tvals[i * KAPPA_SPACE + sumi[kappa][i]] * tvals[i * KAPPA_SPACE + sumi[kappa][i]];
                  }
                  // find maximum sq-sum
                  if (sumtvals[kappa] > maxabst)
                  {
                    maxabst = sumtvals[kappa];
                  }
                }
                // pack the data into the sort helper (kappa and on top the scaled sq-corrcoef-sum such that sorting is for the sq-corrcoef-sum)
                for (size_t kappa = 0; kappa < KAPPA_SPACE*8; kappa++) // each potential kappa
                {
                  sorthelper[kappa] = ((int32_t)(sumtvals[kappa]*(((double)(1<<22))/maxabst))) << 8;
                  sorthelper[kappa] &= 0xffffff00;
                  sorthelper[kappa] |= kappa;
                }
                int32_sort(sorthelper, 8*KAPPA_SPACE); // sort
                for (size_t h = 0; h < 8*KAPPA_SPACE; h++) // find the (known) correct kappa and store its position (=rank)
                {
                  if ((sorthelper[h]&0xff) == ((int32_t)(ckappa[z][x])))
                  {
                    if (h > rank[((x - MINX)*NUMZ + z - MINZ) * (EVAL_NUM / EVOLUTION_RESOLUTION) + count_eval / EVOLUTION_RESOLUTION - 1])
                    {
                      rank[((x - MINX)*NUMZ + z - MINZ) * (EVAL_NUM / EVOLUTION_RESOLUTION) + count_eval / EVOLUTION_RESOLUTION - 1] = h;
                    }
                  }
                } // h
              } // sample
            } // z
          } // x
        } // fi
        
        if (count_eval >= EVAL_NUM)
        {
          break;
        }
      } // end iteration over traces
      if (count_eval >= EVAL_NUM)
      {
        file_offset = it+1;
        break;
      }
    } // end iteration over trace files
        
    // write the ranks
    snprintf(fname, 128, "keccak_snakeranks_%03lu", iterations);
    f = fopen(fname, "w");
    fwrite(rank, sizeof(uint16_t), EVAL_NUM/EVOLUTION_RESOLUTION*NUMX*NUMZ, f);
    fclose(f);
        
    free(traces);
    free(mu);
    free(pt);
    free(Sx);
    free(Sy);
    free(Sx2);
    free(Sy2);
    free(Sxy);
    free(rank);
  } // iterations
}
