# -----------------------------------------------------------------
# COMPANY : Ruhr-Universität Bochum, Chair for Security Engineering
# AUTHOR  : Anna Guinet (email@annagui.net)
# DOCUMENT: A Tale of Snakes and Horses: Amplifying Correlation 
#           Power Analysis on Quadratic Maps, TCHES 2023, Issue 1
# -----------------------------------------------------------------
#
# Python3 script to generate tuples of signal reference values 
# according to Definition 8, for n=3 or n=5, and m=1 or m=n. 

import argparse
import itertools
import sys

def gen_mu_kappa(m_k):
	"""
	Generate all values mu or kappa for m_k bits.

	Parameter:
	m_k -- integer

	Return:
	list of tuples
	"""
	return list(itertools.product([bool(0), bool(1)], repeat=m_k))

def gen_K_prime(m):
	"""
	Generate all values K' for m bits.

	Parameter:
	m -- integer

	Return:
	list of tuples
	"""
	return list(itertools.product([bool(0), bool(1)], repeat=m))

def signal_3D(K_primes, kappas, mus, n, m, m_k):
	"""
	Generate signal power consumption reference values S_ref for all messages mu.

	Parameters:
	K     -- list of bool
	kappa -- list of bool
	mu    -- list of bool
	n     -- integer
	m     -- integer
	m_k   -- integer
	
	Return:
	S -- 3D list of integers, size 2^n x 2^n x 2^n
	"""
	S = []

	for k_prime in range(2**m):
		S_K = []

		for kappa in range(2**m_k):
			S_K_kappa = []

			for mu in range(2**m_k):
				S_K_kappa_mu = int()

				# For each bit i, in each message, for each (K', kappa) 
				for i in range(m):

					# Activity function
					d = K_primes[k_prime][i] ^ kappas[kappa][i] ^ mus[mu][i] ^ (kappas[kappa][(i+1) % n] ^ mus[mu][(i+1) % n] ^ bool(1)) \
						& (kappas[kappa][(i+2) % n] ^ mus[mu][(i+2) % n])

					S_K_kappa_mu += (-1)**(d)

				S_K_kappa.append(S_K_kappa_mu)

			S_K += [S_K_kappa]

		S += [S_K]

	return S

def S_ref(n, m, m_k):
	"""
	Compute signal reference values for chi_n. 
	
	Parameters:
	n   -- integer
	m   -- integer
	m_k -- integer
	"""

	# Secret K'
	K_primes = gen_K_prime(m)

	# Intermediate variables
	mus    = gen_mu_kappa(m_k)
	kappas = gen_mu_kappa(m_k)

	# To display and store K' values
	str_K_primes = []
	for k_prime in K_primes:
		str_K_prime = [str(int(j)) for j in k_prime]
		str_K_prime = ''.join(str_K_prime)
		str_K_primes.append(str_K_prime)
		print("K\'=%s" % str_K_prime)
	print()

	# To display and store kappa values
	str_kappas = []
	for kappa in kappas:
		str_kappa = [str(int(j)) for j in kappa]
		str_kappa = ''.join(str_kappa)
		str_kappas.append(str_kappa)
		print("kappa=%s" % str_kappa)
	print()

	""""""

	# Reference signal power consumption values
	S_ref = signal_3D(K_primes, kappas, mus, n, m, m_k)

	# To display S_ref values
	for S_ref_K_prime in S_ref:
		print("S_ref | K'=%s" % (str_K_primes[S_ref.index(S_ref_K_prime)]))
		for S_ref_kappa in S_ref_K_prime:
			print("kappa=%s: %s" % (str_kappas[S_ref_K_prime.index(S_ref_kappa)], S_ref_kappa))
		print()


def main(unused_command_line_args):
	parser = argparse.ArgumentParser(description='n, m')
	parser.add_argument('n', metavar='n', type=int, default=1, help='length of chi_n, n in {3, 5}')
	parser.add_argument('m', metavar='m', type=int, default=1, help='number of targeted bits m in {1, n}')

	args = parser.parse_args()

	if (args.n != 3) and (args.n != 5):
		print('\n**ERROR**')
		print('Required length of chi output n: 3 or 5 bits')
		print('Wrong value:%s\n' % args.n)

	if (args.m != 1) and (args.m != args.n):
		print('\n**ERROR**')
		print('Required number of targeted bits m: 1 or n bits')
		print('Wrong value: %s\n' % args.m)

	else:
		# Length of S-box
		n = args.n

		# Number of targeted bits
		m = args.m

		if m == n:
			S_ref(n, n, n)
		else:
			S_ref(n, m, 3)

if __name__ == '__main__':
	sys.exit(main(sys.argv))
