
# -----------------------------------------------------------------
# COMPANY : Ruhr-Universität Bochum, Chair for Security Engineering
# AUTHOR  : Anna Guinet (email@annagui.net)
# DOCUMENT: A Tale of Snakes and Horses: Amplifying Correlation 
#           Power Analysis on Quadratic Maps, TCHES 2023, Issue 1
# -----------------------------------------------------------------
#
# Python3 script to generate reduced tuples of signal reference values 
# according to Definition 10, for n = 3 or n = 5. 

import argparse
import itertools
import sys, time

def gen_mu(i, n):
	"""
	Generate all values of mu_(i mod n), mu_(i+1 mod n) and mu_(i+2 mod n).

	Parameters:
	i -- integer in [0,n-1]
	n -- integer
	
	Return:
	mu -- list of tuples
	"""

	# All possible values for 3 bits
	mu = list(itertools.product([bool(0), bool(1)], repeat=3))

	# Transform nested tuples into nested lists
	mu = [list(j) for j in mu]

	# Insert 'False' as the (i+3)-th and (i+4)-th positions
	if n == 5:
		mu_extend = [j.insert((i+3) % n, bool(0)) for j in mu]
		mu_extend = [j.insert((i+4) % n, bool(0)) for j in mu]

	return mu

def gen_key(i, n):
	"""
	Generate all values of key_(i+2 mod n) and key_(i+1 mod n), regardless the i-th bit:
	- key_(i mod n)   = 0
	- key_(i+1 mod n) = kappa_(i+1 mod n)
	- key_(i+2 mod n) = kappa_(i+2 mod n)
	- key_(i+3 mod n) = 0 if n = 5 
	- key_(i+4 mod n) = 0 if n = 5

	Eg: for i = 1, n = 5, key = [[ 0, 0, 0, 0, 0 ], [ 0, 0, 0, 1, 0 ], [ 0, 0, 1, 0, 0 ], [ 0, 0, 1, 1, 0 ]]
	
	Parameters:
	i -- integer in [0,n-1]
	n -- integer
	
	Return:
	key -- list of tuples
	"""
	# All possible values for 2 bits
	key = list(itertools.product([bool(0), bool(1)], repeat=2))

	# Transform nested tuples into nested lists
	key = [list(j) for j in key]

	# Insert 'False' as the i-th, (i+3)-th and (i+4)-th positions
	if n == 5:
		key_extend = [j.insert((i+3) % n, bool(0)) for j in key]
		key_extend = [j.insert((i+4) % n, bool(0)) for j in key]

	# Insert 'False' as the i-th position
	key_extend = [j.insert(i % n, bool(0)) for j in key]
	
	return key
	
def signal_2D(n, i, keys, mus):
	"""
	Generate signal power consumption S for all mu's.

	'key' is a value such as key = kappa, except for bit i: 
	key_i = kappa_i XOR K' = 0

	Parameters:
	n   -- integer
	i   -- integer in [0,n-1]
	key -- list of bool
	mu  -- list of bool

	Return:
	S -- list of integers
	"""
	S = []

	for key in range(len(keys)):
		S_key = []

		for mu in range(2**3):
			# Activity function
			d = keys[key][i] ^ mus[i][mu][i] ^ (keys[key][(i+1) % n] ^ mus[i][mu][(i+1) % n] ^ bool(1)) \
			    & (keys[key][(i+2) % n] ^ mus[i][mu][(i+2) % n])

			# Power consumption model
			S_key_mu = (-1)**(d)
			S_key.append(S_key_mu)

		S += [S_key]
	return S

def reduced_S_ref(n):
	"""
	Compute reduced signal reference values for chi_n. 
	
	Parameter:
	n -- integer
	"""

	# Message mu
	mus = []
	for i in range(n): 
		mu_i = gen_mu(i, n)
		mus.append(mu_i)

	# To display mu values	
	if n==3:
		print("i=0, i=1, i=2") 
		for m in mu_i: # for each mu instance (redundant data in mu to keep the code simpler)
			str_m = [str(int(j)) for j in m]
			str_m = ''.join(str_m)
			print("mu=%s" % str_m)
		print()

	if n==5:
		for mu_i in mus: # for each bit i
			print("i=%s" % mus.index(mu_i)) 
			for m in mu_i: # for each mu instance 
				str_m = [str(int(j)) for j in m]
				str_m = ''.join(str_m)
				print("mu=%s" % str_m)
			print()
	print()


	# Secret key
	keys = []
	for i in range(n): 
		key_i = gen_key(i, n)
		keys.append(key_i)

	# To display key values
	str_keys = []
	for key_i in keys: # for each bit i
		str_key_i = []
		print("i=%s" % keys.index(key_i))
		for k in key_i:
			str_k = [str(int(j)) for j in k]
			str_k = ''.join(str_k)
			str_key_i.append(str_k)
			print("key=%s" % str_k)
		str_keys.append(str_key_i)
		print()
	print()


	# Reduced signal reference values
	S_ref = []
	
	for i in range(n):
		S_ref_i = signal_2D(n, i, keys[i], mus)
		S_ref += [S_ref_i]

		# To display reduced S_ref values
		print("Reduced S_ref | i=%s" % i)
		for s in S_ref_i:
			print("key=%s : %s" % (str_keys[i][S_ref_i.index(s)], s))
		print()

def main(unused_command_line_args):
	parser = argparse.ArgumentParser(description='n')
	parser.add_argument('n', metavar='n', type=int, default=1, help='length of chi_n, n in {3, 5}')

	args = parser.parse_args()

	if (args.n != 3) and (args.n != 5):
		print('\n**ERROR**')
		print('Required length of chi output: 3 or 5 bits')
		print('Wrong value:%s\n' % args.n)

	else:
		# Length of S-box
		n = args.n

		reduced_S_ref(n)

if __name__ == '__main__':
	sys.exit(main(sys.argv))


