/*
 * CHES Artifact for the paper "A Tale of Snakes and Horses: Amplifying Correlation Power Analysis on Quadratic Maps" 
 * (TCHES 2024/1: https://tches.iacr.org/index.php/TCHES/article/view/11239).
 *
 * This program performs the snake attack on Xoodoo.
 */

#include <math.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <immintrin.h>
#include <time.h>
#include "int32_sort.h"

// defines for trace control
#define TRACELEN (42624/2)
#define EVAL_START 13180
#define EVAL_END (EVAL_START + 1)
#define EVALLEN (EVAL_END - EVAL_START)
#define TRACES_PER_FILE 1000
#define TRACES_NUM 10000000
#define EVAL_NUM 75000

#define MAX_ITERATIONS 133
#define EVOLUTION_RESOLUTION 10 // number of traces that are processed before a new ranking is stored

// control coordinates within the state which are attacked
#define MINX 0 // 1-3
#define MAXX 4 // 2-4
#define MINZ 0 // 0-31
#define MAXZ 32 // 1-32
#define NUMX (MAXX-MINX)
#define NUMZ (MAXZ-MINZ)

#define KAPPA_BIT 2
#define KAPPA_SPACE (1<<KAPPA_BIT)
#define M_BIT 3
#define M_SPACE (1<<M_BIT)
#define MU_BIT 3
#define MU_SPACE (1<<MU_BIT)

// helper macros for indexing
#define GETBIT(x, pos, mod) (((x)>>((pos)%(mod)))&1)
#define GETCOLUMN(arr, z, x) (((((arr)[(z + 32*(x + 4*0))/8] >> (z%8)) & 1) << 0) | ((((arr)[(z + 32*(x + 4*1))/8] >> (z%8)) & 1) << 1) | ((((arr)[(z + 32*(x + 4*2))/8] >> (z%8)) & 1) << 2))
#define BIT(x, pos) ((((x)[pos/8])>>((pos)%8))&1)

// helper macros for allocating memory
#define CALLOC8(x, len) do {x = calloc(len, sizeof(uint8_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC16(x, len) do {x = calloc(len, sizeof(int16_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC32(x, len) do {x = calloc(len, sizeof(int32_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)
#define CALLOC64(x, len) do {x = calloc(len, sizeof(int64_t)); if (x == NULL) {printf("bad alloc: %d\n", __LINE__);return -1;}} while(0)

int16_t ref[3][M_SPACE][MU_SPACE][KAPPA_SPACE]; // signal reference vector storage
size_t sumi[8][3] = {{0,0,0},{0,1,2},{2,0,1},{2,1,3},{1,2,0},{1,3,2},{3,2,1},{3,3,3}}; // indices for the snake walk

// precomputation (signal reference vector)
void precompute(void)
{
  uint8_t kappa0,kappa1,kappa2,m,mu;
  for (m = 0; m < M_SPACE; m++)
  {
    for (mu = 0; mu < MU_SPACE; mu++)
    {
      for (kappa1 = 0; kappa1 < 2; kappa1++)
      {
        for (kappa2 = 0; kappa2 < 2; kappa2++)
        {
          int16_t tmp = GETBIT(m, 0, M_BIT) ^ GETBIT(mu, 0, MU_BIT) ^ ((kappa1 ^ GETBIT(mu, 1, MU_BIT) ^ 1) & (kappa2 ^ GETBIT(mu, 2, MU_BIT)));
          ref[0][m][mu][(kappa1 << 1) | kappa2] = 1 - 2*tmp;
        }
      }
    }
  }
  
  for (m = 0; m < M_SPACE; m++)
  {
    for (mu = 0; mu < MU_SPACE; mu++)
    {
      for (kappa2 = 0; kappa2 < 2; kappa2++)
      {
        for (kappa0 = 0; kappa0 < 2; kappa0++)
        {
          int16_t tmp = GETBIT(m, 1, M_BIT) ^ GETBIT(mu, 1, MU_BIT) ^ ((kappa2 ^ GETBIT(mu, 2, MU_BIT) ^ 1) & (kappa0 ^ GETBIT(mu, 0, MU_BIT)));
          ref[1][m][mu][(kappa2 << 1) | kappa0] = 1 - 2*tmp;
        }
      }
    }
  }
  
  for (m = 0; m < M_SPACE; m++)
  {
    for (mu = 0; mu < MU_SPACE; mu++)
    {
      for (kappa0 = 0; kappa0 < 2; kappa0++)
      {
        for (kappa1 = 0; kappa1 < 2; kappa1++)
        {
          int16_t tmp = GETBIT(m, 2, M_BIT) ^ GETBIT(mu, 2, MU_BIT) ^ ((kappa0 ^ GETBIT(mu, 0, MU_BIT) ^ 1) & (kappa1 ^ GETBIT(mu, 1, MU_BIT)));
          ref[2][m][mu][(kappa0 << 1) | kappa1] = 1 - 2*tmp;
        }
      }
    }
  }
}

// correllation coefficient helper function
double corrcoef(int64_t Sxy, int64_t Sx, int64_t Sx2, int64_t Sy, int64_t Sy2, uint32_t num)
{
  int64_t numerator = num*Sxy - (Sx * Sy);
  int64_t X = num*Sx2 - Sx*Sx;
  int64_t Y = num*Sy2 - Sy*Sy;
  double denominator = sqrt(X) * sqrt(Y);
  double result = (double)numerator / denominator;
  if (result < -1 || result > 1)
  {
    #pragma omp critical 
    printf("\nOver- or underflow detected.\n");
  }
  return result;
}

int main(void)
{
  FILE *f;
  char fname[128];
  
  // known kappa for the measurement
  uint8_t correct_kappa[] = {
    0xf3, 0xb5, 0x21, 0x44, 0x27, 0x0c, 0x42, 0x8a, 0xf7, 0x23, 0x0d, 0x3e, 0x47, 0xb4, 0x5c, 0x22, 
    0x49, 0x12, 0xa9, 0x76, 0x51, 0xa2, 0xbb, 0x72, 0x79, 0x58, 0xcd, 0xde, 0xca, 0xb0, 0xe0, 0xe5, 
    0x39, 0x99, 0xa0, 0x34, 0x03, 0xe5, 0x00, 0xd4, 0x58, 0x23, 0x30, 0x19, 0x7e, 0x2a, 0xbd, 0x98
  };
  uint8_t ckappa[32][4] = {0};
  for (size_t z = MINZ; z < MAXZ; z++)
  {
    for (size_t x = MINX; x < MAXX; x++)
    {
      ckappa[z][x] = GETCOLUMN(correct_kappa, z, x);
    } 
  }
  ckappa[3][0] ^= 1; // correct for iota
  ckappa[4][0] ^= 1;
  ckappa[6][0] ^= 1;
  
  // precompute reference vectors
  precompute();
  
  size_t file_offset = 0;
  /////////////////////////////////////////////////////////////////////////////////////// loop over the independent iterations
  for (size_t iterations = 0; iterations < MAX_ITERATIONS; iterations++)
  {
    int16_t *traces;
    uint8_t *mu,*pt;
    int32_t *Sx, *Sy;
    int32_t *Sxy;
    int64_t *Sx2;
    int32_t *Sy2;
    uint8_t rank[EVAL_NUM/EVOLUTION_RESOLUTION*NUMX*NUMZ] = {0};
    CALLOC32(Sx, EVALLEN);
    CALLOC32(Sy, 3*KAPPA_SPACE*NUMX*NUMZ);
    CALLOC32(Sxy, 3*KAPPA_SPACE*EVALLEN*NUMX*NUMZ);
    CALLOC64(Sx2, EVALLEN);
    CALLOC32(Sy2, 3*KAPPA_SPACE*NUMX*NUMZ);
    CALLOC16(traces, TRACELEN*TRACES_PER_FILE);
    CALLOC8(mu, 48*TRACES_PER_FILE);
    CALLOC8(pt, 48*TRACES_PER_FILE);
    
    
    
    ///////////////////////////////////////////////////////////////////////////////////// iteration over trace files
    size_t count_eval = 0;
    for (size_t it = file_offset; it < TRACES_NUM/TRACES_PER_FILE; it++)
    {
      srand(time(NULL));
      
      printf("%4lu / %4d iterations, %10lu / %10d trace files\r", iterations, MAX_ITERATIONS, it, TRACES_NUM/TRACES_PER_FILE);
      fflush(stdout);
      
      // read traces
      snprintf(fname, 128, "traces/trace0/trace0_%d.dat", (int)it);
      f = fopen(fname, "r");
      if (f == NULL || fread(traces, sizeof(int16_t), TRACELEN*TRACES_PER_FILE, f) != TRACELEN*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      for (size_t h = 0; h < TRACELEN*TRACES_PER_FILE; h++)
      {
        traces[h] = ((traces[h]>>8)&0xff) | ((traces[h]<<8)&0xff00);
      }
      
      // read mu
      snprintf(fname, 128, "plaintexts/plain0/mu_%d.dat", (int)it);
      f = fopen(fname, "r");
      if (f == NULL || fread(mu, sizeof(uint8_t), 48*TRACES_PER_FILE, f) != 48*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      // read plaintexts
      snprintf(fname, 128, "plaintexts/plain0/rhoeast_%d.dat", (int)it);
      f = fopen(fname, "r");
      if (f == NULL || fread(pt, sizeof(uint8_t), 48*TRACES_PER_FILE, f) != 48*TRACES_PER_FILE)
      {
        printf("bad fopen: %s\n", fname);
        continue;
      }
      fclose(f);
      f = NULL;
      
      /////////////////////////////////////////////////////////////////////////////////// iteration over traces in one file
      for (size_t j = 0; j < TRACES_PER_FILE; j++)
      {
        for (size_t sample = EVAL_START; sample < EVAL_END; sample++) // each data point in the current trace
        {
          // store sum of points and sum of squared points
          Sx[sample-EVAL_START] += traces[TRACELEN * j + sample];
          Sx2[sample-EVAL_START] += traces[TRACELEN * j + sample] * traces[TRACELEN * j + sample];
        }
  #pragma omp parallel for collapse (2)
        for (int x = MINX; x < MAXX; x++) // iteration over the attack columns
        {
          for (int z = MINZ; z < MAXZ; z++)
          {
            uint64_t thismu = GETCOLUMN(&mu[48*j], z, x);
            uint64_t thispt = GETCOLUMN(&pt[48*j], z, x);
  #pragma omp parallel for collapse (2)
            for (size_t kappa = 0; kappa < KAPPA_SPACE; kappa++) // each potential kappa
            {
              for (size_t i = 0; i < 3; i++)
              {
                int oracle = ref[i][thispt][thismu][kappa];
                // sum of oracle outputs and of their squares
                Sy[(((x - MINX)*NUMZ + z - MINZ)*3 + i) * KAPPA_SPACE + kappa] += oracle;
                Sy2[(((x - MINX)*NUMZ + z - MINZ)*3 + i) * KAPPA_SPACE + kappa] += oracle*oracle;
                for (size_t sample = EVAL_START; sample < EVAL_END; sample++) // each data point in the current trace
                {
                  // sum of oracle times data point
                  Sxy[((((x - MINX)*NUMZ + z - MINZ)*3 + i) * KAPPA_SPACE + kappa) * EVALLEN + sample - EVAL_START] += traces[TRACELEN * j + sample] * oracle;
                }
              } // i
            } // kappa
          } // z
        } // x
        count_eval += 1;
        
        // if evolution resolution is reached, do a ranking snap shot
        // for each attacked row and each sample
        if ((count_eval%EVOLUTION_RESOLUTION) == 0)
        {
  #pragma omp parallel for collapse (3)
          for (int x = MINX; x < MAXX; x++)
          {
            for (int z = MINZ; z < MAXZ; z++)
            {
              for (size_t sample = 0; sample < EVALLEN; sample++)
              {
                double tvals[3*KAPPA_SPACE] = {0}, sumtvals[KAPPA_SPACE*2], maxabst = -2*KAPPA_SPACE;
                int32_t sorthelper[KAPPA_SPACE*2] = {0};
                for (size_t kappa = 0; kappa < KAPPA_SPACE; kappa++) // each potential kappa
                {
                  for (size_t i = 0; i < 3; i++) 
                  {
                    tvals[i * KAPPA_SPACE + kappa] = corrcoef(Sxy[((((x - MINX)*NUMZ + z - MINZ)*3 + i) * KAPPA_SPACE + kappa) * EVALLEN + sample], 
                                    Sx[sample],
                                    Sx2[sample],
                                    Sy[(((x - MINX)*NUMZ + z - MINZ)*3 + i) * KAPPA_SPACE + kappa],
                                    Sy2[(((x - MINX)*NUMZ + z - MINZ)*3 + i) * KAPPA_SPACE + kappa],
				    count_eval);
                  }
                }
                // sum the squared correlation coefficients according to the snake walk
                for (size_t kappa = 0; kappa < 2*KAPPA_SPACE; kappa++)
                {
                  sumtvals[kappa] = 0;
                  for (size_t i = 0; i < 3; i++)
                  {
                    sumtvals[kappa] += tvals[i * KAPPA_SPACE + sumi[kappa][i]] * tvals[i * KAPPA_SPACE + sumi[kappa][i]];
                  }
                  // find maximum sq-sum
                  if (sumtvals[kappa] > maxabst)
                  {
                    maxabst = sumtvals[kappa];
                  }
                }
                // pack the data into the sort helper (kappa and on top the scaled sq-corrcoef-sum such that sorting is for the sq-corrcoef-sum)
                for (size_t kappa = 0; kappa < KAPPA_SPACE*2; kappa++) // each potential kappa
                {
                  sorthelper[kappa] = ((int32_t)(sumtvals[kappa]*(((double)(1<<22))/maxabst))) << 8;
                  sorthelper[kappa] &= 0xffffff00;
                  sorthelper[kappa] |= kappa;
                }
                int32_sort(sorthelper, 2*KAPPA_SPACE); // sort
                for (size_t h = 0; h < 2*KAPPA_SPACE; h++) // find the (known) correct kappa and store its position (=rank)
                {
                  if ((sorthelper[h]&0xff) == ((int32_t)(ckappa[z][x])))
                  {
                    if (h > rank[((x - MINX)*NUMZ + z - MINZ) * (EVAL_NUM / EVOLUTION_RESOLUTION) + count_eval / EVOLUTION_RESOLUTION - 1])
                    {
                      rank[((x - MINX)*NUMZ + z - MINZ) * (EVAL_NUM / EVOLUTION_RESOLUTION) + count_eval / EVOLUTION_RESOLUTION - 1] = h;
                    }
                  }
                } // h
              } // sample
            } // z
          } // x
        } // fi
        
        if (count_eval >= EVAL_NUM)
        {
          break;
        }
      } // end iteration over traces
      if (count_eval >= EVAL_NUM)
      {
        file_offset = it+1;
        break;
      }
    } // end iteration over trace files
    
    // write the ranks
    snprintf(fname, 128, "xoodoo_snakeranks_%03lu", iterations);
    f = fopen(fname, "w");
    fwrite(rank, sizeof(uint8_t), EVAL_NUM/EVOLUTION_RESOLUTION*NUMX*NUMZ, f);
    fclose(f);
        
    free(traces);
    free(mu);
    free(pt);
    free(Sx);
    free(Sy);
    free(Sx2);
    free(Sy2);
    free(Sxy);
  } // iterations
}
